@@ -117,6 +117,20 @@ func verifyNodePointers(t *testing.T, n *Node) {
 	testTrue(t, parent == nil || parent.LastChild == cur)
 }
 
+func TestChildNodes(t *testing.T) {
+	t.Run("Has 3 children", func(t *testing.T) {
+		node := &Node{Type: ElementNode}
+
+		AddChild(node, &Node{Type: CommentNode})
+		AddChild(node, &Node{Type: ElementNode})
+		AddChild(node, &Node{Type: TextNode})
+
+		children := node.ChildNodes()
+
+		testTrue(t, len(children) == 3)
+	})
+}
+
 func TestAddAttr(t *testing.T) {
 	for _, test := range []struct {
 		name     string
@@ -124,6 +138,7 @@ func TestAddAttr(t *testing.T) {
 		key      string
 		val      string
 		expected string
+		ok       bool
 	}{
 		{
 			name:     "node has no existing attr",
@@ -131,6 +146,7 @@ func TestAddAttr(t *testing.T) {
 			key:      "ns:k1",
 			val:      "v1",
 			expected: `< ns:k1="v1"></>`,
+			ok:       true,
 		},
 		{
 			name:     "node has existing attrs",
@@ -138,10 +154,20 @@ func TestAddAttr(t *testing.T) {
 			key:      "k2",
 			val:      "v2",
 			expected: `< k1="v1" k2="v2"></>`,
+			ok:       true,
+		},
+		{
+			name:     "node already has existing attr",
+			n:        &Node{Type: AttributeNode, Attr: []Attr{{Name: xml.Name{Local: "k1"}, Value: "v1"}}},
+			key:      "k1",
+			val:      "v1",
+			expected: `< k1="v1"></>`,
+			ok:       false,
 		},
 	} {
 		t.Run(test.name, func(t *testing.T) {
-			AddAttr(test.n, test.key, test.val)
+			ok := AddAttr(test.n, test.key, test.val)
+			testTrue(t, ok == test.ok)
 			testValue(t, test.n.OutputXML(true), test.expected)
 		})
 	}
@@ -154,6 +180,7 @@ func TestSetAttr(t *testing.T) {
 		key      string
 		val      string
 		expected string
+		ok       bool
 	}{
 		{
 			name:     "node has no existing attr",
@@ -161,6 +188,7 @@ func TestSetAttr(t *testing.T) {
 			key:      "ns:k1",
 			val:      "v1",
 			expected: `< ns:k1="v1"></>`,
+			ok:       true,
 		},
 		{
 			name:     "node has an existing attr, overwriting",
@@ -168,6 +196,7 @@ func TestSetAttr(t *testing.T) {
 			key:      "ns:k1",
 			val:      "v2",
 			expected: `< ns:k1="v2"></>`,
+			ok:       true,
 		},
 		{
 			name:     "node has no existing attr, no ns",
@@ -175,6 +204,7 @@ func TestSetAttr(t *testing.T) {
 			key:      "k1",
 			val:      "v1",
 			expected: `< k1="v1"></>`,
+			ok:       true,
 		},
 		{
 			name:     "node has an existing attr, no ns, overwriting",
@@ -182,11 +212,13 @@ func TestSetAttr(t *testing.T) {
 			key:      "k1",
 			val:      "v2",
 			expected: `< k1="v2"></>`,
+			ok:       true,
 		},
 	} {
 
 		t.Run(test.name, func(t *testing.T) {
-			test.n.SetAttr(test.key, test.val)
+			ok := test.n.SetAttr(test.key, test.val)
+			testTrue(t, ok == test.ok)
 			testValue(t, test.n.OutputXML(true), test.expected)
 		})
 	}
@@ -198,48 +230,54 @@ func TestRemoveAttr(t *testing.T) {
 		n        *Node
 		key      string
 		expected string
+		ok       bool
 	}{
 		{
 			name:     "node has no existing attr",
 			n:        &Node{Type: AttributeNode},
 			key:      "ns:k1",
 			expected: `<></>`,
+			ok:       false,
 		},
 		{
 			name:     "node has an existing attr, overwriting",
 			n:        &Node{Type: AttributeNode, Attr: []Attr{{Name: xml.Name{Space: "ns", Local: "k1"}, Value: "v1"}}},
 			key:      "ns:k1",
 			expected: `<></>`,
+			ok:       true,
 		},
 		{
 			name:     "node has no existing attr, no ns",
 			n:        &Node{Type: AttributeNode},
 			key:      "k1",
 			expected: `<></>`,
+			ok:       false,
 		},
 		{
 			name:     "node has an existing attr, no ns, overwriting",
 			n:        &Node{Type: AttributeNode, Attr: []Attr{{Name: xml.Name{Local: "k1"}, Value: "v1"}}},
 			key:      "k1",
 			expected: `<></>`,
+			ok:       true,
 		},
 	} {
 
 		t.Run(test.name, func(t *testing.T) {
-			test.n.RemoveAttr(test.key)
+			ok := test.n.RemoveAttr(test.key)
+			testTrue(t, ok == test.ok)
 			testValue(t, test.n.OutputXML(true), test.expected)
 		})
 	}
 }
 
 func TestRemoveFromTree(t *testing.T) {
-	xml := `<?procinst?>
+	xmlStr := `<?procinst?>
 		<!--comment-->
 		<aaa><bbb/>
 			<ddd><eee><fff/></eee></ddd>
 		<ggg/></aaa>`
 	parseXML := func() *Node {
-		doc, err := Parse(strings.NewReader(xml))
+		doc, err := Parse(strings.NewReader(xmlStr))
 		testTrue(t, err == nil)
 		return doc
 	}
@@ -250,7 +288,7 @@ func TestRemoveFromTree(t *testing.T) {
 		testTrue(t, n != nil)
 		RemoveFromTree(n)
 		verifyNodePointers(t, doc)
-		testValue(t, doc.OutputXML(false),
+		testValue(t, doc.OutputXMLWithOptions(WithoutPreserveSpace()),
 			`<?procinst?><!--comment--><aaa><bbb></bbb><ddd></ddd><ggg></ggg></aaa>`)
 	})
 
@@ -260,7 +298,7 @@ func TestRemoveFromTree(t *testing.T) {
 		testTrue(t, n != nil)
 		RemoveFromTree(n)
 		verifyNodePointers(t, doc)
-		testValue(t, doc.OutputXML(false),
+		testValue(t, doc.OutputXMLWithOptions(WithoutPreserveSpace()),
 			`<?procinst?><!--comment--><aaa><ddd><eee><fff></fff></eee></ddd><ggg></ggg></aaa>`)
 	})
 
@@ -270,7 +308,7 @@ func TestRemoveFromTree(t *testing.T) {
 		testTrue(t, n != nil)
 		RemoveFromTree(n)
 		verifyNodePointers(t, doc)
-		testValue(t, doc.OutputXML(false),
+		testValue(t, doc.OutputXMLWithOptions(WithoutPreserveSpace()),
 			`<?procinst?><!--comment--><aaa><bbb></bbb><ggg></ggg></aaa>`)
 	})
 
@@ -280,17 +318,17 @@ func TestRemoveFromTree(t *testing.T) {
 		testTrue(t, n != nil)
 		RemoveFromTree(n)
 		verifyNodePointers(t, doc)
-		testValue(t, doc.OutputXML(false),
+		testValue(t, doc.OutputXMLWithOptions(WithoutPreserveSpace()),
 			`<?procinst?><!--comment--><aaa><bbb></bbb><ddd><eee><fff></fff></eee></ddd></aaa>`)
 	})
 
 	t.Run("remove decl node works", func(t *testing.T) {
 		doc := parseXML()
 		procInst := doc.FirstChild
-		testValue(t, procInst.Type, DeclarationNode)
+		testValue(t, procInst.Type, ProcessingInstruction)
 		RemoveFromTree(procInst)
 		verifyNodePointers(t, doc)
-		testValue(t, doc.OutputXML(false),
+		testValue(t, doc.OutputXMLWithOptions(WithoutPreserveSpace()),
 			`<!--comment--><aaa><bbb></bbb><ddd><eee><fff></fff></eee></ddd><ggg></ggg></aaa>`)
 	})
 
@@ -300,7 +338,7 @@ func TestRemoveFromTree(t *testing.T) {
 		testValue(t, commentNode.Type, CommentNode)
 		RemoveFromTree(commentNode)
 		verifyNodePointers(t, doc)
-		testValue(t, doc.OutputXML(false),
+		testValue(t, doc.OutputXMLWithOptions(WithoutPreserveSpace()),
 			`<?procinst?><aaa><bbb></bbb><ddd><eee><fff></fff></eee></ddd><ggg></ggg></aaa>`)
 	})
 
@@ -308,11 +346,36 @@ func TestRemoveFromTree(t *testing.T) {
 		doc := parseXML()
 		RemoveFromTree(doc)
 		verifyNodePointers(t, doc)
-		testValue(t, doc.OutputXML(false),
+		testValue(t, doc.OutputXMLWithOptions(WithoutPreserveSpace()),
 			`<?procinst?><!--comment--><aaa><bbb></bbb><ddd><eee><fff></fff></eee></ddd><ggg></ggg></aaa>`)
 	})
 }
 
+func TestAddImmediateSibling(t *testing.T) {
+	s := `<?xml version="1.0" encoding="UTF-8"?>
+    <AAA>
+        <BBB id="1"/>
+        <CCC id="2">
+            <DDD/>
+        </CCC>
+		<CCC id="3">
+            <DDD/>
+        </CCC>
+     </AAA>`
+	root, err := Parse(strings.NewReader(s))
+	if err != nil {
+		t.Error(err)
+	}
+
+	aaa := findNode(root, "AAA")
+	n := aaa.SelectElement("BBB")
+	if n == nil {
+		t.Fatalf("n is nil")
+	}
+	AddImmediateSibling(n, &Node{Type: ElementNode, Data: "r"})
+	testValue(t, root.OutputXMLWithOptions(WithoutPreserveSpace()), `<?xml version="1.0" encoding="UTF-8"?><AAA><BBB id="1"></BBB><r></r><CCC id="2"><DDD></DDD></CCC><CCC id="3"><DDD></DDD></CCC></AAA>`)
+}
+
 func TestSelectElement(t *testing.T) {
 	s := `<?xml version="1.0" encoding="UTF-8"?>
     <AAA>
@@ -373,7 +436,8 @@ func TestEscapeValueWrite(t *testing.T) {
 	}
 
 	var b strings.Builder
-	root.Write(&b, true)
+	err = root.Write(&b, true)
+	testTrue(t, err == nil)
 	escapedInnerText := b.String()
 	if !strings.Contains(escapedInnerText, "&lt;*&gt;") {
 		t.Fatal("Inner Text has not been escaped")
@@ -423,7 +487,8 @@ func TestUnnecessaryEscapeValueWrite(t *testing.T) {
 	}
 
 	var b strings.Builder
-	root.Write(&b, true)
+	err = root.Write(&b, true)
+	testTrue(t, err == nil)
 	escapedInnerText := b.String()
 	if strings.Contains(escapedInnerText, "&#x9") {
 		t.Fatal("\\n has been escaped unnecessarily")
@@ -467,7 +532,8 @@ func TestHtmlUnescapeStringOriginStringWrite(t *testing.T) {
 	}
 
 	var b strings.Builder
-	root.Write(&b, false)
+	err = root.Write(&b, false)
+	testTrue(t, err == nil)
 	escapedInnerText := b.String()
 	unescapeString := html.UnescapeString(escapedInnerText)
 	if strings.Contains(unescapeString, "&amp;") {
@@ -491,13 +557,13 @@ func TestWriteWithNamespacePrefix(t *testing.T) {
 	s := `<?xml version="1.0" encoding="UTF-8"?><S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/"><S:Body></S:Body></S:Envelope>`
 	doc, _ := Parse(strings.NewReader(s))
 	var b strings.Builder
-	doc.Write(&b, false)
+	err := doc.Write(&b, false)
+	testTrue(t, err == nil)
 	if s != b.String() {
 		t.Fatal("xml document missing some characters")
 	}
 }
 
-
 func TestQueryWithPrefix(t *testing.T) {
 	s := `<?xml version="1.0" encoding="UTF-8"?><S:Envelope xmlns:S="http://schemas.xmlsoap.org/soap/envelope/"><S:Body test="1"><ns2:Fault xmlns:ns2="http://schemas.xmlsoap.org/soap/envelope/" xmlns:ns3="http://www.w3.org/2003/05/soap-envelope"><faultcode>ns2:Client</faultcode><faultstring>This is a client fault</faultstring></ns2:Fault></S:Body></S:Envelope>`
 	doc, _ := Parse(strings.NewReader(s))
@@ -582,7 +648,7 @@ func TestOutputXMLWithSpaceDirect(t *testing.T) {
 		t.Errorf(`expected "%s", obtained "%s"`, expected, g)
 	}
 
-	output := html.UnescapeString(doc.OutputXML(true))
+	output := html.UnescapeString(doc.OutputXMLWithOptions(WithOutputSelf(), WithoutPreserveSpace()))
 	if strings.Contains(output, "\n") {
 		t.Errorf("the outputted xml contains newlines")
 	}
@@ -606,7 +672,7 @@ func TestOutputXMLWithSpaceOverwrittenToPreserve(t *testing.T) {
 		t.Errorf(`expected "%s", obtained "%s"`, expected, g)
 	}
 
-	output := html.UnescapeString(doc.OutputXML(true))
+	output := html.UnescapeString(doc.OutputXMLWithOptions(WithOutputSelf(), WithoutPreserveSpace()))
 	if strings.Contains(output, "\n") {
 		t.Errorf("the outputted xml contains newlines")
 	}
@@ -680,8 +746,8 @@ func TestOutputXMLWithPreserveSpaceOption(t *testing.T) {
 		</student>
 	</class_list>`
 	doc, _ := Parse(strings.NewReader(s))
-	resultWithSpace := doc.OutputXMLWithOptions(WithPreserveSpace())
-	resultWithoutSpace := doc.OutputXMLWithOptions()
+	resultWithSpace := doc.OutputXMLWithOptions()
+	resultWithoutSpace := doc.OutputXMLWithOptions(WithoutPreserveSpace())
 	if !strings.Contains(resultWithSpace, "> Robert <") {
 		t.Errorf("output was not expected. expected %v but got %v", " Robert ", resultWithSpace)
 	}
@@ -746,3 +812,39 @@ func TestOutputXMLWithSingleQuotes(t *testing.T) {
 		t.Errorf(`expected "%s", obtained "%s"`, expected, output)
 	}
 }
+
+func TestProcessingInstruction(t *testing.T) {
+	xml := `<?ProcInstTag random string ?><?AnotherProcInst a="b"?><a/>`
+
+	doc, err := Parse(strings.NewReader(xml))
+	if err != nil {
+		t.Fatalf("Parse failed: %v", err)
+	}
+
+	firstNode := doc.FirstChild
+	if firstNode.Type != ProcessingInstruction {
+		t.Errorf("First node type should be ProcessingInstruction, got %d", firstNode.Type)
+	}
+
+	piNode := FindOne(doc, "//ProcInstTag")
+	if piNode == nil {
+		t.Fatal("FindOne should be find one, but got nil")
+	}
+	if piNode.ProcInst == nil {
+		t.Fatal("ProcInstData is nil")
+	}
+
+	if piNode.ProcInst.Target != "ProcInstTag" {
+		t.Errorf("ProcInstData.Target mismatch: expected 'ProcInstTag', got '%s'", piNode.ProcInst.Target)
+	}
+
+	if piNode.ProcInst.Inst != "random string" {
+		t.Errorf("ProcInstData.Inst mismatch: expected 'random string', got '%s'", piNode.ProcInst.Inst)
+	}
+
+	output := doc.OutputXML(true)
+	expected := `<?ProcInstTag random string?><?AnotherProcInst a="b"?><a></a>`
+	if output != expected {
+		t.Errorf("Output mismatch:\nExpected: %s\nGot: %s", expected, output)
+	}
+}
