@@ -2,6 +2,7 @@ package xmlquery
 
 import (
 	"bufio"
+	"bytes"
 	"encoding/xml"
 	"fmt"
 	"io"
@@ -37,17 +38,58 @@ func Parse(r io.Reader) (*Node, error) {
 
 // ParseWithOptions is like parse, but with custom options
 func ParseWithOptions(r io.Reader, options ParserOptions) (*Node, error) {
-	p := createParser(r)
-	options.apply(p)
-	for {
-		_, err := p.parse()
-		if err == io.EOF {
-			return p.doc, nil
-		}
+	var lineStarts []int
+	// If line numbers are requested, read all data for position tracking
+	if options.WithLineNumbers {
+		var err error
+		var data []byte
+		data, err = io.ReadAll(r)
 		if err != nil {
 			return nil, err
 		}
+		r = bytes.NewReader(data)
+
+		// Pre-calculate line starts
+		lineStarts = []int{0}
+		for i, b := range data {
+			if b == '\n' {
+				lineStarts = append(lineStarts, i+1)
+			}
+		}
 	}
+
+	p := createParser(r)
+	if options.WithLineNumbers {
+		p.lineStarts = lineStarts
+	}
+	options.apply(p)
+
+	var err error
+	for err == nil {
+		_, err = p.parse()
+	}
+
+	if err == io.EOF {
+		// additional check for validity
+		// according to: https://www.w3.org/TR/xml
+		// the document MUST contain at least ONE element
+		valid := false
+		for doc := p.doc; doc != nil; doc = doc.NextSibling {
+			for node := doc.FirstChild; node != nil; node = node.NextSibling {
+				if node.Type == ElementNode {
+					valid = true
+					break
+				}
+			}
+		}
+		if !valid {
+			return nil, fmt.Errorf("xmlquery: invalid XML document")
+		}
+
+		return p.doc, nil
+	}
+
+	return nil, err
 }
 
 type parser struct {
@@ -62,6 +104,10 @@ type parser struct {
 	reader              *cachedReader // Need to maintain a reference to the reader, so we can determine whether a node contains CDATA.
 	once                sync.Once
 	space2prefix        map[string]*xmlnsPrefix
+	currentLine         int // Track current line number during parsing
+	lastProcessedPos    int // Track how much cached data we've already processed for line counting
+
+	lineStarts []int
 }
 
 type xmlnsPrefix struct {
@@ -72,10 +118,13 @@ type xmlnsPrefix struct {
 func createParser(r io.Reader) *parser {
 	reader := newCachedReader(bufio.NewReader(r))
 	p := &parser{
-		decoder: xml.NewDecoder(reader),
-		doc:     &Node{Type: DocumentNode},
-		level:   0,
-		reader:  reader,
+		decoder:          xml.NewDecoder(reader),
+		doc:              &Node{Type: DocumentNode},
+		level:            0,
+		reader:           reader,
+		currentLine:      0,
+		lastProcessedPos: 0,
+		lineStarts:       nil,
 	}
 	if p.decoder.CharsetReader == nil {
 		p.decoder.CharsetReader = charset.NewReaderLabel
@@ -84,6 +133,37 @@ func createParser(r io.Reader) *parser {
 	return p
 }
 
+// updateLineNumber scans only new cached data for newlines to update current line position
+func (p *parser) updateLineNumber() {
+	if p.lineStarts == nil {
+		return
+	}
+	offset := int(p.decoder.InputOffset())
+	for i := p.currentLine; i < len(p.lineStarts); i++ {
+		if offset > p.lineStarts[i] && p.lineStarts[i] >= p.lastProcessedPos {
+			p.currentLine = i + 1
+			break
+		}
+		if offset <= p.lineStarts[i] {
+			break
+		}
+	}
+	p.lastProcessedPos = offset
+	/*
+		cached := p.reader.CacheWithLimit(-1) // Get all cached data
+
+		// Only process data we haven't seen before
+		for i := p.lastProcessedPos; i < len(cached); i++ {
+			if cached[i] == '\n' {
+				p.currentLine++
+			}
+		}
+
+		// Update our position to avoid reprocessing this data
+		p.lastProcessedPos = len(cached)
+	*/
+}
+
 func (p *parser) parse() (*Node, error) {
 	p.once.Do(func() {
 		p.space2prefix = map[string]*xmlnsPrefix{"http://www.w3.org/XML/1998/namespace": {name: "xml", level: 0}}
@@ -94,6 +174,10 @@ func (p *parser) parse() (*Node, error) {
 		p.reader.StartCaching()
 		tok, err := p.decoder.Token()
 		p.reader.StopCaching()
+
+		// Update line number based on processed content
+		p.updateLineNumber()
+
 		if err != nil {
 			return nil, err
 		}
@@ -106,10 +190,11 @@ func (p *parser) parse() (*Node, error) {
 				attributes[0].Name = xml.Name{Local: "version"}
 				attributes[0].Value = "1.0"
 				node := &Node{
-					Type:  DeclarationNode,
-					Data:  "xml",
-					Attr:  attributes,
-					level: 1,
+					Type:       DeclarationNode,
+					Data:       "xml",
+					Attr:       attributes,
+					level:      1,
+					LineNumber: p.currentLine,
 				}
 				AddChild(p.prev, node)
 				p.level = 1
@@ -153,6 +238,7 @@ func (p *parser) parse() (*Node, error) {
 				NamespaceURI: tok.Name.Space,
 				Attr:         attributes,
 				level:        p.level,
+				LineNumber:   p.currentLine,
 			}
 
 			if p.level == p.prev.level {
@@ -168,7 +254,7 @@ func (p *parser) parse() (*Node, error) {
 
 			if node.NamespaceURI != "" {
 				if v, ok := p.space2prefix[node.NamespaceURI]; ok {
-					cached := string(p.reader.Cache())
+					cached := string(p.reader.CacheWithLimit(len(v.name) + len(node.Data) + 2))
 					if strings.HasPrefix(cached, fmt.Sprintf("%s:%s", v.name, node.Data)) || strings.HasPrefix(cached, fmt.Sprintf("<%s:%s", v.name, node.Data)) {
 						node.Prefix = v.name
 					}
@@ -228,13 +314,12 @@ func (p *parser) parse() (*Node, error) {
 			}
 		case xml.CharData:
 			// First, normalize the cache...
-			cached := strings.ToUpper(string(p.reader.Cache()))
+			cached := bytes.ToUpper(p.reader.CacheWithLimit(9))
 			nodeType := TextNode
-			if strings.HasPrefix(cached, "<![CDATA[") || strings.HasPrefix(cached, "![CDATA[") {
+			if bytes.HasPrefix(cached, []byte("<![CDATA[")) || bytes.HasPrefix(cached, []byte("![CDATA[")) {
 				nodeType = CharDataNode
 			}
-
-			node := &Node{Type: nodeType, Data: string(tok), level: p.level}
+			node := &Node{Type: nodeType, Data: string(tok), level: p.level, LineNumber: p.currentLine}
 			if p.level == p.prev.level {
 				AddSibling(p.prev, node)
 			} else if p.level > p.prev.level {
@@ -246,7 +331,7 @@ func (p *parser) parse() (*Node, error) {
 				AddSibling(p.prev.Parent, node)
 			}
 		case xml.Comment:
-			node := &Node{Type: CommentNode, Data: string(tok), level: p.level}
+			node := &Node{Type: CommentNode, Data: string(tok), level: p.level, LineNumber: p.currentLine}
 			if p.level == p.prev.level {
 				AddSibling(p.prev, node)
 			} else if p.level > p.prev.level {
@@ -258,10 +343,11 @@ func (p *parser) parse() (*Node, error) {
 				AddSibling(p.prev.Parent, node)
 			}
 		case xml.ProcInst: // Processing Instruction
-			if p.prev.Type != DeclarationNode {
-				p.level++
+			level := p.level
+			if p.prev.Type == DocumentNode {
+				level = p.level + 1
 			}
-			node := &Node{Type: DeclarationNode, Data: tok.Target, level: p.level}
+			node := &Node{Type: DeclarationNode, Data: tok.Target, level: level, LineNumber: p.currentLine}
 			pairs := strings.Split(string(tok.Inst), " ")
 			for _, pair := range pairs {
 				pair = strings.TrimSpace(pair)
@@ -269,19 +355,24 @@ func (p *parser) parse() (*Node, error) {
 					AddAttr(node, pair[:i], strings.Trim(pair[i+1:], `"'`))
 				}
 			}
-			if p.level == p.prev.level {
+			if tok.Target != "xml" {
+				node.Type = ProcessingInstruction
+				node.ProcInst = &ProcInstData{Target: tok.Target, Inst: strings.TrimSpace(string(tok.Inst))}
+			}
+			if level == p.prev.level {
 				AddSibling(p.prev, node)
-			} else if p.level > p.prev.level {
+			} else if level > p.prev.level {
 				AddChild(p.prev, node)
-			} else if p.level < p.prev.level {
-				for i := p.prev.level - p.level; i > 1; i-- {
+			} else if level < p.prev.level {
+				for i := p.prev.level - level; i > 1; i-- {
 					p.prev = p.prev.Parent
 				}
 				AddSibling(p.prev.Parent, node)
 			}
 			p.prev = node
+			p.level = level
 		case xml.Directive:
-			node := &Node{Type: NotationNode, Data: string(tok), level: p.level}
+			node := &Node{Type: NotationNode, Data: string(tok), level: p.level, LineNumber: p.currentLine}
 			if p.level == p.prev.level {
 				AddSibling(p.prev, node)
 			} else if p.level > p.prev.level {
